//    Copyright (C) 2010  Luis Alfonso Arce González, ajaest[@]gmail.com
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

package net.ajaest.core.main;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import javax.swing.UIManager;

import net.ajaest.core.exporters.Exporter;
import net.ajaest.core.main.ExceptionHandler.KNOWN_EXCEPTIONS;
import net.ajaest.data.kanji.KanjiGraph;
import net.ajaest.data.kanji.KanjiStroke;
import net.ajaest.data.kanji.KanjiTag;
import net.ajaest.data.kanji.SimilarKanjiStrokeDemuxer;

//TODO: JavaDoc
public class Launcher {

	public static void main(String... args) {
		/*
		 * TODO: WARNING! i-bus and OpenJDK doesn't get along very well, it's a
		 * bug(I don't know whether is a OpenJDK bug or IBUS bug). Button + dead
		 * keys won't return, for example, accented chars in Spanish's or
		 * French's keyboards. This problem is solved if i-bus input method is
		 * deactivated or if the jvm is initialized with 「XMODIFIERS=''」
		 * parameter, but you probably won't be able to write in japanese.
		 */
		
		// Sets system's look and feel
		Messages.localePrintln("**Launcher: initializing system look and feel");
		try {
				UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());

		} catch (Exception e) {
			ExceptionHandler.handleException(e, KNOWN_EXCEPTIONS.LOOK_AND_FEEL_DOES_NOT_EXISTS);
		}

		Messages.localePrintln("**Launcher: system look and feel initialized");

		// default config file
		JDKOptions jdkOpt = new JDKOptions();

		// Now we have info about language, initializing
		Messages.localePrintln("**Launcher: initializing messages system");
		Locale.setDefault(new Locale(jdkOpt.getLanguage()));
		Messages.initMessages(jdkOpt.getLanguage());
		Messages.localePrintln("**Launcher: messages system initialized");

		// Retrieve exporters
		Messages.localePrintln("**Launcher: retrieving exporters");
		List<Exporter<KanjiTag>> exporters = retrieveExporters("exporters/");
		if (exporters != null) {
			for (Exporter<KanjiTag> exp : exporters) {
				exp.setLanguaje(jdkOpt.getLanguage());
			}
		} else {
			new File("exporters" + File.separator).mkdirs();
			exporters = new ArrayList<Exporter<KanjiTag>>();
		}
		Messages.localePrintln("**Launcher: exporters retrieved");
		
		// initializing javadikt but not running it
		Messages.localePrintln("**Launcher: initializing JavaDiKt core");
		JavaDiKt jdk = new JavaDiKt(jdkOpt);
		jdk.setExporters(exporters);
		try {
			jdk.setSimilarStrokeDemuxer(createSimilarKanjiStrokeDemuxer());
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		Messages.localePrintln("**Launcher: JavaDiKt core initialized");

		// now exception handler can catch javadikt core exceptions
		Messages.localePrintln("**Launcher: initializing exception handler");
		ExceptionHandler.setJavaDiKt(jdk);
		Messages.localePrintln("**Launcher: exception handler initialized");
		
		// running javadikt
		Messages.localePrintln("**Launcher: running javadikt");
		jdk.run();
	}

	/**
	 * 
	 * @param path
	 *            The directory path that contains the exporters
	 * @return A list of exporters in the specified file path. The list will be
	 *         empty if there aren't exporters in the specified path. {@code
	 *         null} will be returned if the path doesn't exits or if it doesn't
	 *         denotes a directory.
	 */
	@SuppressWarnings("unchecked")
	private static List<Exporter<KanjiTag>> retrieveExporters(String path) {

		List<Exporter<KanjiTag>> exporters;
		File exportersPath = new File(path);

		if (exportersPath.exists() && exportersPath.isDirectory()) {
			
			exporters =  new ArrayList<Exporter<KanjiTag>>();
			
			for (File exp : exportersPath.listFiles()) {
				if (exp.isFile() && exp.getName().endsWith(".obj")) {
					try {

						ObjectInputStream oos = new ObjectInputStream(new FileInputStream("exporters/textExporter.obj"));
						Object ret = oos.readObject();
						oos.close();
						if (ret instanceof Exporter<?>)
							exporters.add((Exporter<KanjiTag>) ret);

					} catch (FileNotFoundException e) {
						e.printStackTrace();
					} catch (IOException e) {
						e.printStackTrace();
					} catch (ClassNotFoundException e) {
						e.printStackTrace();
					}
				}
			}
		} else {
			exporters = null;
		}

		return exporters;
	}

	@SuppressWarnings("unchecked")
	private static SimilarKanjiStrokeDemuxer createSimilarKanjiStrokeDemuxer() throws FileNotFoundException, IOException, ClassNotFoundException {

		ObjectInputStream ois = new ObjectInputStream(new FileInputStream("dict/subgraphset.obj"));
		Set<List<KanjiStroke>> allowedsubGraphs = (Set<List<KanjiStroke>>) ois.readObject();
		ois.close();

		ois = new ObjectInputStream(new FileInputStream("dict/strokemap.obj"));
		Map<KanjiStroke, Set<KanjiStroke>> similarityMap = (Map<KanjiStroke, Set<KanjiStroke>>) ois.readObject();
		ois.close();

		ois = new ObjectInputStream(new FileInputStream("dict/graphset.obj"));
		Set<KanjiGraph> allowedGraph = (Set<KanjiGraph>) ois.readObject();
		ois.close();

		SimilarKanjiStrokeDemuxer sksd = new SimilarKanjiStrokeDemuxer(similarityMap, allowedsubGraphs, allowedGraph);

		return sksd;

	}
}
