//    Copyright (C) 2010  Luis Alfonso Arce González, ajaest[@]gmail.com
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

package net.ajaest.data.auxi;

/**
 * 
 * The {@code UByte} class wraps a value of primitive type {@code byte} in an
 * object and provides several methods to manipulate it as unsigned byte number.
 * The use of this class is preferred for storing purposes.
 * 
 * @author Luis Alfonso Arce González
 * @see java.lang.Number
 * @see java.lang.Byte
 */
public class UByte extends Number implements Comparable<UByte> {

	private static final long serialVersionUID = -5487850224168814313L;

	private byte value;

	public static final UByte MIN_VALUE = new UByte(0);
	public static final UByte MAX_VALUE = new UByte(255);

	public UByte(byte b) {
		this.value = b;
	}

	public UByte(int ui) {
		this((byte) (ui % 256));
	}

	public UByte(String s) {
		this(Byte.parseByte(s));
	}

	/**
	 * Returns a UByte whose value is {@code (this + val)}.
	 * 
	 * @param val
	 *            value to be added to this UByte.
	 * @return {@code this + val}
	 */
	public UByte add(UByte val) {

		return new UByte(this.intValue() + val.intValue());
	}

	/**
	 * Returns a UByte whose value is {@code abs(this - val)}.
	 * 
	 * @param val
	 *            value to be subtracted from this UByte.
	 * @return {@code this - val}
	 */
	public UByte subtract(UByte val) {

		return new UByte(Math.abs(this.intValue() - val.intValue()));
	}

	/**
	 * Returns a UByte whose value is {@code (this * val)}.
	 * 
	 * @param val
	 *            value to be multiplied by this UByte.
	 * @return {@code this * val}
	 */
	public UByte multiply(UByte val) {

		return new UByte(this.intValue() * val.intValue());
	}

	/**
	 * Returns a UByte whose value is {@code (this<sup>2</sup>)}.
	 * 
	 * @return {@code this<sup>2</sup>}
	 */
	public UByte square() {

		return new UByte(this.intValue() * this.intValue());
	}

	/**
	 * Returns a UByte whose value is {@code (this / val)}.
	 * 
	 * @param val
	 *            value by which this UByte is to be divided.
	 * @return {@code this / val}
	 * @throws ArithmeticException
	 *             {@code val==0}
	 */
	public UByte divide(UByte val) {

		return new UByte(this.intValue() / val.intValue());
	}

	/**
	 * Returns a UByte whose value is {@code (this mod m}).
	 * 
	 * @param m
	 *            the modulus.
	 * @return {@code this mod m}
	 */
	public UByte mod(UByte m) {

		return new UByte(this.intValue() % m.intValue());
	}

	/**
	 * Returns a UByte whose value is <tt>(this<sup>exponent</sup>)</tt>. Note
	 * that {@code exponent} is an integer rather than a UByte.
	 * 
	 * @param exponent
	 *            exponent to which this UByte is to be raised.
	 * @return <tt>this<sup>exponent</sup></tt>
	 * @throws ArithmeticException
	 *             {@code exponent} is negative. (This would cause the operation
	 *             to yield a non-integer value.)
	 */
	public UByte pow(int exponent) {

		return new UByte((int) Math.pow(this.intValue(), exponent));
	}

	/**
	 * Returns a UByte whose value is the greatest common divisor of {@code
	 * abs(this)} and {@code abs(val)}. Returns 0 if {@code val==0}.
	 * 
	 * @param val
	 *            value with which the GCD is to be computed.
	 * @return {@code GCD(this, val)}
	 */
	public UByte gcd(UByte val) {

		int i2 = val.intValue();

		if (i2 != 0) {

			int i1 = this.intValue();
			int temp = i1 % i2;

			while (temp != 0) {
				i1 = i2;
				i2 = temp;
				temp = i1 % i2;
			}
		}

		return new UByte(i2);
	}

	@Override
	public byte byteValue() {

		return value;
	}

	@Override
	public double doubleValue() {

		return (double) intValue();
	}

	@Override
	public float floatValue() {

		return (float) intValue();
	}

	@Override
	public int intValue() {

		return ((int) value) & 0xFF;
	}

	@Override
	public long longValue() {

		return (long) intValue();
	}

	@Override
	public int compareTo(UByte o) {

		return this.intValue() - o.intValue();
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + value;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		UByte other = (UByte) obj;
		if (value != other.value)
			return false;
		return true;
	}

	@Override
	public String toString() {

		return "" + intValue();
	}

}
