//    Copyright (C) 2010  Luis Alfonso Arce González, ajaest[@]gmail.com
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

package net.ajaest.data.dict;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import net.ajaest.data.auxi.OrderedSequenceTree;
import net.ajaest.data.auxi.database.DatabaseManager;
import net.ajaest.data.auxi.database.QAbout;
import net.ajaest.data.dict.query.IntegerValueQAbout;
import net.ajaest.data.dict.query.KanjiExpression;
import net.ajaest.data.dict.query.StringValueQAbout;
import net.ajaest.data.dict.query.ValueQAbout;
import net.ajaest.data.dict.query.kanjiFields.ClassicNelsonQAbout;
import net.ajaest.data.dict.query.kanjiFields.ClassicRadicalQAbout;
import net.ajaest.data.dict.query.kanjiFields.CommonStrokeMiscountQAbout;
import net.ajaest.data.dict.query.kanjiFields.DeRooQAbout;
import net.ajaest.data.dict.query.kanjiFields.DicNameQAbout;
import net.ajaest.data.dict.query.kanjiFields.DicReferenceQAbout;
import net.ajaest.data.dict.query.kanjiFields.FourCornerQAbout;
import net.ajaest.data.dict.query.kanjiFields.FrequencyQAbout;
import net.ajaest.data.dict.query.kanjiFields.GradeQAbout;
import net.ajaest.data.dict.query.kanjiFields.JISCharsetQAbout;
import net.ajaest.data.dict.query.kanjiFields.JISCodeQAbout;
import net.ajaest.data.dict.query.kanjiFields.JLPTLevelQAbout;
import net.ajaest.data.dict.query.kanjiFields.KanjiGraphQAbout;
import net.ajaest.data.dict.query.kanjiFields.MeaningLanguageQAbout;
import net.ajaest.data.dict.query.kanjiFields.MeaningQAbout;
import net.ajaest.data.dict.query.kanjiFields.ReadingQAbout;
import net.ajaest.data.dict.query.kanjiFields.ReadingTypeQAbout;
import net.ajaest.data.dict.query.kanjiFields.SkipQAbout;
import net.ajaest.data.dict.query.kanjiFields.SpahnHadamitzkyQAbout;
import net.ajaest.data.dict.query.kanjiFields.StrokeCountQAbout;
import net.ajaest.data.dict.query.kanjiFields.UnicodeValueQAbout;
import net.ajaest.data.dict.query.kanjiFields.VariantReferenceQAbout;
import net.ajaest.data.dict.query.kanjiFields.VariantTypeQAbout;
import net.ajaest.data.kanji.AllowedStrokeLineEnum;
import net.ajaest.data.kanji.DicReferencePair;
import net.ajaest.data.kanji.Kanji;
import net.ajaest.data.kanji.KanjiGraph;
import net.ajaest.data.kanji.KanjiReference;
import net.ajaest.data.kanji.KanjiStroke;
import net.ajaest.data.kanji.KanjiTag;
import net.ajaest.data.kanji.MeaningEntry;
import net.ajaest.data.kanji.Radical;
import net.ajaest.data.kanji.ReadingEntry;
import net.ajaest.data.kanji.VariantPair;
import net.ajaest.tools.parsers.StrokeOrderParser;

import org.neodatis.odb.CorruptedDatabaseException;
import org.neodatis.odb.ODB;
import org.neodatis.odb.ODBFactory;
import org.neodatis.odb.Objects;
import org.neodatis.odb.core.NeoDatisError;
import org.neodatis.odb.core.query.criteria.And;
import org.neodatis.odb.core.query.criteria.ICriterion;
import org.neodatis.odb.core.query.criteria.Or;
import org.neodatis.odb.core.query.criteria.Where;
import org.neodatis.odb.impl.core.query.criteria.CriteriaQuery;

/**
 * This class provides methods to manage a kanji database.
 * 
 * @author Luis Alfonso Arce González
 * 
 */
// I've created a monster!
public class NeodatisKanjiDatabaseManager implements DatabaseManager<KanjiTag, KanjiExpression> {

	private ODB kanjiDb;
	private KanjiDatabaseInfo kdbi;
	private OrderedSequenceTree<AllowedStrokeLineEnum, KanjiGraph> graphTree;
	private Map<Integer, SortedSet<Radical>> radicals;

	private static final Map<Class<?>, String> fieldNamesTranslations = new HashMap<Class<?>, String>();
	static {
		// Object with a unique instance per kanji, query from nested value in
		// kanji
		fieldNamesTranslations.put(UnicodeValueQAbout.class, "unicode");

		fieldNamesTranslations.put(JISCharsetQAbout.class, "jisCode.first");
		fieldNamesTranslations.put(JISCodeQAbout.class, "jisCode.second");

		fieldNamesTranslations.put(ClassicRadicalQAbout.class, "classicRadical");

		fieldNamesTranslations.put(ClassicNelsonQAbout.class, "nelsonRadical");

		fieldNamesTranslations.put(GradeQAbout.class, "grade");

		fieldNamesTranslations.put(StrokeCountQAbout.class, "strokeCount");

		fieldNamesTranslations.put(CommonStrokeMiscountQAbout.class, "strokeMiscounts");

		fieldNamesTranslations.put(FrequencyQAbout.class, "frequency");

		fieldNamesTranslations.put(JLPTLevelQAbout.class, "JLPTLevel");

		fieldNamesTranslations.put(SkipQAbout.class, "queryCodes.skip");
		fieldNamesTranslations.put(SpahnHadamitzkyQAbout.class, "queryCodes.spahnHadamitzky");
		fieldNamesTranslations.put(DeRooQAbout.class, "queryCodes.deRoo");
		fieldNamesTranslations.put(FourCornerQAbout.class, "queryCodes.fourCorner");

		// List value object, query from StringPair/DicEntry specific inherited
		// class

		fieldNamesTranslations.put(VariantReferenceQAbout.class, "second");
		fieldNamesTranslations.put(VariantTypeQAbout.class, "first");

		fieldNamesTranslations.put(DicNameQAbout.class, "first");
		fieldNamesTranslations.put(DicReferenceQAbout.class, "second");

		fieldNamesTranslations.put(ReadingTypeQAbout.class, "key");
		fieldNamesTranslations.put(ReadingQAbout.class, "elements");

		fieldNamesTranslations.put(MeaningLanguageQAbout.class, "key");
		fieldNamesTranslations.put(MeaningQAbout.class, "elements");;
	}

	/**
	 * Constructs a kanji database manager interface and connects it to the
	 * kanji database in {@code path}
	 * 
	 * @param f
	 *            Kanji database path {@code File} object.
	 * @throws IOException
	 */
	public NeodatisKanjiDatabaseManager(File kanjiDatabase, File strokeOrder) throws IOException {
		this(kanjiDatabase.getAbsolutePath(), strokeOrder.getAbsolutePath());
	}

	/**
	 * Constructs a kanji database manager interface and connects it to the
	 * kanji database in {@code path}
	 * 
	 * @param kanjiDatabase
	 *            Kanji database path.
	 * @param string
	 * @throws IOException
	 */
	public NeodatisKanjiDatabaseManager(String kanjiDatabasePath, String strokeOrderPath) throws IOException {

		kanjiDb = ODBFactory.open(kanjiDatabasePath);

		Objects<KanjiDatabaseInfo> kdbi = kanjiDb.getObjects(KanjiDatabaseInfo.class);

		buildGraphTree(strokeOrderPath);
		
		buildRadicalMap();

		if (kdbi.size() != 1)
			throw new CorruptedDatabaseException(NeoDatisError.UNEXPECTED_SITUATION, "There are multiples data headers or does not exists anyone.");

		this.kdbi = kdbi.getFirst();
	}

	private void buildRadicalMap() {

		radicals = new HashMap<Integer, SortedSet<Radical>>();

		Objects<Radical> retRadicals = kanjiDb.getObjects(Radical.class);

		SortedSet<Radical> temp;
		for (Radical r : retRadicals) {
			temp = radicals.get(r.radicalNumber);
			if (temp == null) {
				temp = new TreeSet<Radical>();
				radicals.put(r.getRadicalNumber(), temp);
			}
			temp.add(r);
		}	
	}

	/**
	 * Closes the connection between the database manager interface and the
	 * kanji database.
	 */
	@Override
	public void close() {
		kanjiDb.close();
		graphTree = null;
	}

	/**
	 * Executes a kanji expression representing a kanji query in order to
	 * retrieve the kanjis in database that fulfills the specified conditions.
	 * 
	 * @param query
	 *            The {@code KanjiExpression} representing a kanji query.
	 * @return A set of kanjis as a result of executing the query.
	 */
	@Override
	public Set<KanjiTag> executeQuery(KanjiExpression query) {

		KanjiExpression optimized = optimizeQuery(query);

		// Object with a unique instance per kanji, query from nested value in
		// kanji
		List<And> aboutKanji = new ArrayList<And>();
		// List value object, query from StringPair/DicEntry specific inherited
		// class
		List<And> aboutDicReference = new ArrayList<And>();
		List<And> aboutReadings = new ArrayList<And>();
		List<And> aboutMeaning = new ArrayList<And>();
		List<And> aboutVariant = new ArrayList<And>();
		List<List<KanjiGraph>> aboutGraph = new ArrayList<List<KanjiGraph>>();
		List<List<KanjiExpression>> aboutQueryIntoQuery = new ArrayList<List<KanjiExpression>>();

		int size = optimized.size();

		for (int i = 0; i < size; i++) {

			switch (optimized.getBooleanCases().get(i)) {
				case AND :
					translateCriteria(optimized.getDomains().get(i), aboutKanji, aboutDicReference, aboutReadings, aboutMeaning, aboutVariant, aboutGraph, aboutQueryIntoQuery);
					break;
				case OR :
				case FIRST :

					aboutKanji.add(Where.and());
					aboutDicReference.add(Where.and());
					aboutReadings.add(Where.and());
					aboutMeaning.add(Where.and());
					aboutVariant.add(Where.and());
					aboutGraph.add(new ArrayList<KanjiGraph>());
					aboutQueryIntoQuery.add(new ArrayList<KanjiExpression>());

					translateCriteria(optimized.getDomains().get(i), aboutKanji, aboutDicReference, aboutReadings, aboutMeaning, aboutVariant, aboutGraph, aboutQueryIntoQuery);
					break;

				default :
					throw new RuntimeException("Boolean operator not supported.");
			}
		}

		return executeQuery(aboutKanji, aboutDicReference, aboutReadings, aboutMeaning, aboutVariant, aboutGraph, aboutQueryIntoQuery);
	}

	private List<AllowedStrokeLineEnum> getLineSequence(KanjiGraph kg) {

		List<AllowedStrokeLineEnum> sequence = new ArrayList<AllowedStrokeLineEnum>();

		for (KanjiStroke ks : kg.getStrokes()) {
			for (AllowedStrokeLineEnum asl : ks.getStrokeList()) {

				sequence.add(asl);
			}
		}

		return sequence;
	}

	private Set<KanjiTag> executeQuery(List<And> aboutKanji, List<And> aboutDicReference, List<And> aboutReadings, List<And> aboutMeaning, List<And> aboutVariant, List<List<KanjiGraph>> aboutGraph, List<List<KanjiExpression>> aboutQueryIntoQuery) {

		// partial current and results
		Set<KanjiTag> andResult = null;
		//partial current graph result
		Set<KanjiReference> graphResult = null;
		//final or result
		Set<KanjiTag> orResult = new HashSet<KanjiTag>();
		// stores non kanji results
		
		
		CriteriaQuery tempQuery = null;
		// all lists must have the same size
		int size = aboutDicReference.size();
		boolean execute = size>0;
		
		// A lot of improvements could be done here. For now, we only want it to
		// work.
		for (int i = 0; i < size /* && execute */; i++) {

			if (!aboutKanji.get(i).isEmpty() && execute) {
				tempQuery = new CriteriaQuery(Kanji.class, aboutKanji.get(i));
				andResult = executeNeodatisQuery(tempQuery);
				execute = andResult.size() > 0;
			}

			if (!aboutDicReference.get(i).isEmpty() && execute) {
				andResult = executeNeodatisQuery(DicReferencePair.class, aboutDicReference.get(i), andResult);
				execute = andResult.size() > 0;
			}

			if (!aboutReadings.get(i).isEmpty() && execute) {
				andResult = executeNeodatisQuery(ReadingEntry.class, aboutReadings.get(i), andResult);
				execute = andResult.size() > 0;
			}

			if (!aboutMeaning.get(i).isEmpty() && execute) {
				andResult = executeNeodatisQuery(MeaningEntry.class, aboutMeaning.get(i), andResult);
				execute = andResult.size() > 0;
			}

			if (!aboutVariant.get(i).isEmpty() && execute) {
				andResult = executeNeodatisQuery(VariantPair.class, aboutVariant.get(i), andResult);
				execute = andResult.size() > 0;
			}

			if (!aboutGraph.get(i).isEmpty() && execute) {
				
				List<AllowedStrokeLineEnum> tempAsl = null;
				for (int j = 0; j < aboutGraph.get(i).size() && execute; j++) {
					KanjiGraph kg = aboutGraph.get(i).get(j);
					tempAsl = getLineSequence(kg);
					graphResult = new HashSet<KanjiReference>(this.graphTree.search(tempAsl));
					
					if (andResult == null)
						andResult = getKanjis(graphResult);
					else
						andResult.retainAll(getKanjis(graphResult));
					execute = andResult.size() > 0;
				}
			}

			if (!aboutQueryIntoQuery.get(i).isEmpty() && execute) {

				for (int j = 0; j < aboutQueryIntoQuery.get(i).size() && execute; j++) {
					KanjiExpression ke = aboutQueryIntoQuery.get(i).get(j);

					if (andResult == null)
						andResult = executeQuery(ke);
					else
						andResult.retainAll(executeQuery(ke));

					execute = andResult.size() > 0;
				}
			}

			orResult.addAll(andResult);
			andResult = null;
			// TODO: remove
			execute = true;
			// --
		}

		

		return orResult;
	}
	
	private And nonKanjiQueryCriteriaShortener(Set<KanjiTag> previousQuery, And and) {

		Or or = null;
		And ret = null;

		if (previousQuery == null || previousQuery.size() == 0) {
			ret = and;

		} else {
			or = Where.or();
			ret = Where.and();
			// makes retain all operation form previous query using unicode
			// references
			for (KanjiTag kt : previousQuery)
				or.add(Where.equal("unicodeRef", kt.getUnicode()));

			ret.add(or);
			ret.add(and);
		}
		
		return ret;
	}

	private <E> Set<E> executeNeodatisQuery(CriteriaQuery qc) {

		Objects<E> obs = kanjiDb.getObjects(qc);

		return new HashSet<E>(obs);
	}
	
	private Set<KanjiTag> executeNeodatisQuery(Class<?> c, And and, Set<KanjiTag> previousQuery) {

		And quickAnd = nonKanjiQueryCriteriaShortener(previousQuery, and);
		CriteriaQuery tempQuery = new CriteriaQuery(c, quickAnd);

		Set<KanjiReference> partialResult = executeNeodatisQuery(tempQuery);

		Set<KanjiTag> result = partialResult.size() > 0 ? getKanjis(partialResult) : new HashSet<KanjiTag>();

		return result;
	}

	private Set<KanjiTag> getKanjis(Set<KanjiReference> kr) {

		Or or = Where.or();

		for (KanjiReference obj : kr) {
			or.add(Where.equal("unicode", obj.getUnicodeRef()));
		}

		Objects<KanjiTag> obs = kanjiDb.getObjects(new CriteriaQuery(Kanji.class, or));

		return new HashSet<KanjiTag>(obs);

	}

	@Override
	protected void finalize() throws Throwable {
		this.kanjiDb.close();
		super.finalize();
	}

	/**
	 * Returns the current kanji database information object.
	 * 
	 * @return a {@code KanjiDatabaseInformation} object.
	 */
	public KanjiDatabaseInfo getKanjiDatabaseInformation() {
		return kdbi;
	}

	/**
	 * Returns the current kanji database stroke order sequence tree.
	 * 
	 * @return the kanji stroke order tree
	 */
	public OrderedSequenceTree<AllowedStrokeLineEnum, KanjiGraph> getStrokeTree() {
		return graphTree;
	}

	private KanjiExpression optimizeQuery(KanjiExpression ke) {
		// TODO:Dummy, implement. I need performance tests before going on it.
		return ke;
	}

	private ICriterion translateCaseToNeodatis(IntegerValueQAbout qa) {

		switch (qa.getQueryCase()) {
			case EQUALS :
				return Where.equal(translateFieldName(qa), qa.getValue());
			case GREATER_OR_EQUALS_THAN :
				return Where.ge(translateFieldName(qa), qa.getValue());
			case GREATER_THAN :
				return Where.gt(translateFieldName(qa), qa.getValue());
			case LESS_OR_EQUALS_THAN :
				return Where.le(translateFieldName(qa), qa.getValue());
			case LESS_THAN :
				return Where.lt(translateFieldName(qa), qa.getValue());
			case NOT_EQUALS :
				return Where.not(Where.equal(translateFieldName(qa), qa.getValue()));
			case NULL :
				return Where.isNull(translateFieldName(qa));

			default :
				throw new IllegalArgumentException("Query case not suportted");
		}
	}

	private ICriterion translateCaseToNeodatis(StringValueQAbout qa) {

		switch (qa.getQueryCase()) {
			case EQUALS :
				return Where.equal(translateFieldName(qa), qa.getValue());
			case GREATER_OR_EQUALS_THAN :
				return Where.ge(translateFieldName(qa), qa.getValue());
			case GREATER_THAN :
				return Where.gt(translateFieldName(qa), qa.getValue());
			case LESS_OR_EQUALS_THAN :
				return Where.le(translateFieldName(qa), qa.getValue());
			case LESS_THAN :
				return Where.lt(translateFieldName(qa), qa.getValue());
			case NOT_EQUALS :
				return Where.not(Where.equal(translateFieldName(qa), qa.getValue()));
			case NULL :
				return Where.isNull(translateFieldName(qa));

			default :
				throw new IllegalArgumentException("Query case not suportted");
		}
	}

	private void translateCriteria(QAbout qa, List<And> aboutKanji, List<And> aboutDicReference, List<And> aboutReadings, List<And> aboutMeaning, List<And> aboutVariant, List<List<KanjiGraph>> aboutGraph, List<List<KanjiExpression>> aboutQueryIntoQuery) {
		// TODO: repair
		int last = aboutGraph.size() - 1;// all list must have the same size

		// query about Kanji class/kanji nested class, IntegerValueQAbout, not
		// list
		if (qa instanceof ClassicNelsonQAbout || qa instanceof ClassicRadicalQAbout || qa instanceof FrequencyQAbout || qa instanceof GradeQAbout || qa instanceof JLPTLevelQAbout || qa instanceof StrokeCountQAbout || qa instanceof UnicodeValueQAbout) {

			aboutKanji.get(last).add(translateCaseToNeodatis((IntegerValueQAbout) qa));

		}
		// query about Kanji class/kanji nested class, StringValueQAbout, not
		// list
		else if (qa instanceof DeRooQAbout || qa instanceof FourCornerQAbout || qa instanceof SkipQAbout || qa instanceof SpahnHadamitzkyQAbout) {

			aboutKanji.get(last).add(translateCaseToNeodatis((StringValueQAbout) qa));

		}
		// query about Kanji class/kanji nested class, StringValueQAbout, not
		// list
		else if (qa instanceof JISCharsetQAbout || qa instanceof JISCodeQAbout) {

			aboutKanji.get(last).add(translateCaseToNeodatis((StringValueQAbout) qa));

		}
		// query about Kanji class, IntegerValueQAbout, list
		else if (qa instanceof CommonStrokeMiscountQAbout) {

			aboutKanji.get(last).add(Where.contain(translateFieldName((IntegerValueQAbout) qa), ((IntegerValueQAbout) qa).getValue()));
			// TODO: WARNING: this means that always a case about
			// miscounts will be "equals"
		}
		// Query about graph
		else if (qa instanceof KanjiGraphQAbout) {
			aboutGraph.get(last).add(((KanjiGraphQAbout) qa).getValue());
		}
		// query about StringEntry classes, meaning
		else if (qa instanceof MeaningLanguageQAbout) {

			aboutMeaning.get(last).add(translateCaseToNeodatis((StringValueQAbout) qa));
		}
		// query about StringEntry classes, meaning entry
		else if (qa instanceof MeaningQAbout) {

			aboutMeaning.get(last).add(Where.contain(translateFieldName((StringValueQAbout) qa), ((StringValueQAbout) qa).getValue()));
			// TODO: WARNING: this means that always a case about
			// miscounts will be "equals"
		}
		// query about StringEntry classes, reading key
		else if (qa instanceof ReadingQAbout) {

			aboutReadings.get(last).add(Where.contain(translateFieldName((StringValueQAbout) qa), ((StringValueQAbout) qa).getValue()));
			// TODO: WARNING: this means that always a case about
			// miscounts will be "equals"

		}
		// query about StringEntry classes, reading entry
		else if (qa instanceof ReadingTypeQAbout) {

			aboutReadings.get(last).add(translateCaseToNeodatis((StringValueQAbout) qa));

		}
		// query about StringPair classes, variant
		else if (qa instanceof VariantReferenceQAbout || qa instanceof VariantTypeQAbout) {

			aboutVariant.get(last).add(translateCaseToNeodatis((StringValueQAbout) qa));
		}
		// query about StringPair classes, dictionary reference
		else if (qa instanceof DicNameQAbout || qa instanceof DicReferenceQAbout) {

			aboutDicReference.get(last).add(translateCaseToNeodatis((StringValueQAbout) qa));
		}
		// query into query
		else if (qa instanceof KanjiExpression) {
			aboutQueryIntoQuery.get(last).add((KanjiExpression) qa);
		}

		else {
			throw new RuntimeException("Domain not supported.");
		}
	}

	private void buildGraphTree(String path) throws IOException {
		
		StrokeOrderParser sop = new StrokeOrderParser(path, false);
		sop.parse();
		sop.close();

		graphTree = new OrderedSequenceTree<AllowedStrokeLineEnum, KanjiGraph>(false);

		for (Integer i : sop.getRetrieved().keySet()) {
			graphTree.add(getLineSequence(sop.getRetrieved().get(i)), sop.getRetrieved().get(i));
		}

		for (Integer i : sop.getRepeated().keySet()) {
			graphTree.add(getLineSequence(sop.getRepeated().get(i)), sop.getRetrieved().get(i));
		}

	}

	public ODB getLowLevelDatabaseManager() {
		return this.kanjiDb;
	}

	private String translateFieldName(ValueQAbout<?> qa) {

		return fieldNamesTranslations.get(qa.getClass());
	}
	
	public SortedSet<Radical> getRadicalInfos(Integer radNumber) {
		
		return radicals.get(radNumber);
	}

}
